/*
 *  Copyright (C) 2012  Marian Vlcek, Jakub Cizek
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *  If you publish any work which uses this program and/or data
 *  produced by this program, please cite following paper: 
 * 
 *  J. Cizek. M. Vlcek, I. Prochazka, Digital spectrometer
 *  for coincidence measurement of Doppler broadening of positron
 *  annihilation radiation, Nuclear Instruments and Methods
 *  in Physics Research, Section A 623, 982-994 (2010)
 */

#include "stdafx.h"
#include <stdio.h>
#include <math.h>
#include "histogram.h"

#define pi 3.1415926535897932384626433

Window::Window(double (*func)(int,int), int l) : len(l), win(new double[len]) {
	int i;
	double sum;
	sum=0;
	for (i = 0; i < len; i++) {
		win[i]=func(i, len);
		sum+=win[i];
	}
	for (i = 0; i < len; i++) {
		win[i]=win[i]*1.0/sum;
	}
}

inline double coef(double x) {
	return cos(pi*x*0.5)*cos(pi*x*0.5);
}

void interpolate(double val, int *r, int *g, int *b, struct color a[], const int n) {
	int lo, hi;
	val=val*(n-1);
	lo = (int)floor(val);
	hi = (int)ceil(val);

	if (lo < 0) lo = 0;
	if (lo > n-1) lo = n-1;

	if (hi < 0) hi = 0;
	if (hi > n-1) hi = n-1;

	*r = (int)((coef(hi-val)*(a[hi]).r+coef(val-lo)*(a[lo]).r)*1.0/(coef(val-lo)+coef(hi-val)));
	*g = (int)((coef(hi-val)*(a[hi]).g+coef(val-lo)*(a[lo]).g)*1.0/(coef(val-lo)+coef(hi-val)));
	*b = (int)((coef(hi-val)*(a[hi]).b+coef(val-lo)*(a[lo]).b)*1.0/(coef(val-lo)+coef(hi-val)));
	
}

void histogram1D::smooth(const Window window) {
	int i,j,k;
	double sum;
	double *orig_data = new double[bins];
 

	if (bins < window.len) {
		printf("Data length must be larger than window length!\n");
		return;
	}
	for (int c = 0; c < channels; c++) { 
		for (i = 0; i < bins; i++) {
			orig_data[i]=hist[i][c];
			hist[i][c]=0;
		}
		for (i = 1; i < bins; i++) {
			sum=0;
			for (j = -window.len/2; j <= window.len/2; j++) {
				k=i+j;
				if (k < 1) continue;
				if (k >= bins - 1) continue;
				hist[i][c]+=orig_data[k]*window.win[j+window.len/2];
				sum+=window.win[j+window.len/2];
			}
			hist[i][c]=hist[i][c]/sum;
		}
	}
	delete[] orig_data;
}



histogram1D::histogram1D(int b, int c, double l, double u)
: bins(b), channels(c), lower(l), upper(u), hist(boost::extents[bins][channels])
{
  for(int i = 0; i < bins; i++) 
    for(int j = 0; j < channels; j++)
        hist[i][j] = 0;    
}

histogram1D::~histogram1D()
{
}


void histogram1D::add(const double val, const int c)
{
	int index;
	index = (int) (bins - 1) * (val - lower) * 1.0 / (upper - lower) + 0.5;
	if (index >= bins) index = bins - 1;
	if (index < 0) index = 0;
	hist[index][c]++;    
}

void histogram1D::clearData()
{
	for (int i=0; i<bins; i++) {
		for (int j=0; j<channels; j++) {
			hist[i][j]=0;
		}
	}
}

void histogram1D::write(const char path[])
{
	FILE *outfile;
	outfile = fopen(path,"w");
	for (int i=0; i<bins; i++) {
		fprintf(outfile, "%f", lower + i * 1.0 / (bins - 1) * (upper - lower));
		for (int j=0; j<channels; j++) {
			fprintf(outfile, " %f", hist[i][j]);
		}
		fprintf(outfile, "\n");
	}	
	fclose(outfile);	
}

void histogram1D::calibrate(const double a, const double b)
{
	lower = a * lower + b;
	upper = a * upper + b;
}

double histogram1D::fwhm(const double val, const int c)
{
	if (val < lower || val > upper) return 0;
	int index, i_max, i_lo, i_hi;
	double lo, hi;
	index = (int) (bins - 1) * (val - lower) * 1.0 / (upper - lower) + 0.5;
	i_max=index;
	for (int i = index - 2; i < index + 2; i++) {
		if (hist[i][c] > hist[i_max][c]) i_max=i;
	}
// 	printf("max = %g, pos = %d\n", hist[i_max][c], i_max);
	i_lo=i_max;
	i_hi=i_max;
	while (hist[i_lo][c] > 0.5*hist[i_max][c]) i_lo--;
	while (hist[i_hi][c] > 0.5*hist[i_max][c]) i_hi++;
	lo = (hist[i_lo+1][c] - 0.5*hist[i_max][c])*i_lo + (0.5*hist[i_max][c] - hist[i_lo][c])*(i_lo+1);
	hi = (hist[i_hi-1][c] - 0.5*hist[i_max][c])*i_hi + (0.5*hist[i_max][c] - hist[i_hi][c])*(i_hi-1);
	lo /= hist[i_lo+1][c] - hist[i_lo][c];
	hi /= hist[i_hi-1][c] - hist[i_hi][c];
// 	printf("i_lo = %d, i_hi = %d\n", i_lo, i_hi);
// 	printf("lo = %g, hi = %g\n", lo, hi);
// 	printf("FWHM = %g\n", (hi - lo) * 1.0 / (bins - 1) * (upper - lower));
	return (hi - lo) * 1.0 / (bins - 1) * (upper - lower);
}

histogram2D::histogram2D(const int x_b, const int y_b, const int c, const double x_l, const double x_u, const double y_l, const double y_u)
: x_bins(x_b), y_bins(y_b), channels(c), x_lower(x_l), x_upper(x_u), y_lower(y_l), y_upper(y_u), hist(boost::extents[x_bins][y_bins][channels])
{
	for(int i = 0; i < x_bins; i++) 
		for(int j = 0; j < y_bins; j++)
			for(int k = 0; k < channels; k++)
				hist[i][j][k] = 0;    
}

histogram2D::~histogram2D()
{
}


void histogram2D::add(const double x_val, const double y_val, const int c)
{
	int x_index, y_index;
	x_index = (int) (x_bins - 1) * (x_val - x_lower) * 1.0 / (x_upper - x_lower) + 0.5;
	if (x_index >= x_bins) x_index = x_bins - 1;
	if (x_index < 0) x_index = 0;
	y_index = (int) (y_bins - 1) * (y_val - y_lower) * 1.0 / (y_upper - y_lower) + 0.5;
	if (y_index >= y_bins) y_index = y_bins - 1;
	if (y_index < 0) y_index = 0;	
	hist[x_index][y_index][c]++;    
}

void histogram2D::writeText(const char path[])
{
	FILE *outfile;
	outfile = fopen(path,"w");
	for (int i=0; i<x_bins; i++) {
		for (int j=0; j<y_bins; j++) {
			fprintf(outfile, "%f", x_lower + i * 1.0 / (x_bins - 1) * (x_upper - x_lower));
			fprintf(outfile, " %f", y_lower + j * 1.0 / (y_bins - 1) * (y_upper - y_lower));
			for (int k=0; k<channels; k++) {
				fprintf(outfile, " %f", hist[i][j][k]);
			}
			fprintf(outfile, "\n");			
		}
	}	
	fclose(outfile);	
}

void histogram2D::writeConstXCut(const char path[], const int x, const bool normalize)
{
	double sum[2], coef[2];
	FILE *outfile;
	outfile = fopen(path,"w");
	for (int k=0; k<channels; k++) {
		sum[k]=0;
		coef[k]=1;
	}
	if (normalize == true) {
		for (int j=0; j<y_bins; j++) {
			for (int k=0; k<channels; k++) {
				sum[k]+=hist[x][j][k];
			}
		}
		for (int k=0; k<channels; k++) {
			coef[k]=1e6/sum[k];
		}
	}
	for (int j=0; j<y_bins; j++) {
		fprintf(outfile, " %f", y_lower + j * 1.0 / (y_bins - 1) * (y_upper - y_lower));
		for (int k=0; k<channels; k++) {
			fprintf(outfile, " %f", hist[x][j][k]*coef[k]);
		}
		fprintf(outfile, "\n");			
	}	
	fclose(outfile);	
}

void histogram2D::writeConstYCut(const char path[], const int y, const bool normalize)
{
	double sum[2], coef[2];
	FILE *outfile;
	outfile = fopen(path,"w");
	for (int k=0; k<channels; k++) {
		sum[k]=0;
		coef[k]=1;
	}
	if (normalize == true) {
		for (int i=0; i<y_bins; i++) {
			for (int k=0; k<channels; k++) {
				sum[k]+=hist[i][y][k];
			}
		}
		for (int k=0; k<channels; k++) {
			coef[k]=1e6/sum[k];
		}
	}
	for (int i=0; i<x_bins; i++) {
		fprintf(outfile, "%f", x_lower + i * 1.0 / (x_bins - 1) * (x_upper - x_lower));
		for (int k=0; k<channels; k++) {
			fprintf(outfile, " %f", hist[i][y][k]*coef[k]);
		}
		fprintf(outfile, "\n");			
	}	
	fclose(outfile);	
}

double histogram2D::fwhmConstXCutWide(const int x, const double val, const int w, const int c)
{
	double *cut = new double[y_bins];
	int index, i_max, i_lo, i_hi;
	double lo, hi;
	if (val < y_lower || val > y_upper) {
		delete[] cut;
		return 0;
	}
	for (int j=0; j<y_bins; j++) {
		cut[j]=0;
		for (int n = -w; n <= w; n++) {
			cut[j]+=hist[x+n][j][c];
		}
	}
	index = (int) (y_bins - 1) * (val - y_lower) * 1.0 / (y_upper - y_lower) + 0.5;
	i_max=index;
	for (int i = index - 2; i < index + 2; i++) {
		if (cut[i] > cut[i_max]) i_max=i;
	}

	i_lo=i_max;
	i_hi=i_max;
	while (cut[i_lo] > 0.5*cut[i_max]) i_lo--;
	while (cut[i_hi] > 0.5*cut[i_max]) i_hi++;
	lo = (cut[i_lo+1] - 0.5*cut[i_max])*i_lo + (0.5*cut[i_max] - cut[i_lo])*(i_lo+1);
	hi = (cut[i_hi-1] - 0.5*cut[i_max])*i_hi + (0.5*cut[i_max] - cut[i_hi])*(i_hi-1);
	lo /= cut[i_lo+1] - cut[i_lo];
	hi /= cut[i_hi-1] - cut[i_hi];

	delete[] cut;
	return (hi - lo) * 1.0 / (y_bins - 1) * (y_upper - y_lower);
}

void histogram2D::writeConstXCutWide(const char path[], const int x, const int w, const bool normalize)
{
	double *sum = new double[channels];
	double *coef = new double[channels];
	FILE *outfile;
	outfile = fopen(path,"w");
	double ** cut = 0;
	cut = new double*[y_bins];
	for (int j=0; j<y_bins; j++) {
		cut[j] = new double[channels];
	}
	for (int k=0; k<channels; k++) {
		sum[k]=0;
		coef[k]=1;
		for (int j=1; j<y_bins-1; j++) {
			cut[j][k]=0;
			for (int n = -w; n <= w; n++) {
				cut[j][k]+=hist[x+n][j][k];
			}
			sum[k]+=cut[j][k];
		}
		if (normalize == true) {
			coef[k]=1e6/sum[k];
		}
	}
	for (int j=0; j<y_bins; j++) {
		fprintf(outfile, " %f", y_lower + j * 1.0 / (y_bins - 1) * (y_upper - y_lower));
		for (int k=0; k<channels; k++) {
			fprintf(outfile, " %f", cut[j][k]*coef[k]);
		}
		fprintf(outfile, "\n");			
	}
	for (int j=y_bins; j>0; --j) {
		delete[] cut[j-1];
	}
	delete[] cut;
	fclose(outfile);	
}

void histogram2D::writeConstYCutWide(const char path[], const int y, const int w, const bool normalize)
{
	double sum[2], coef[2];
	FILE *outfile;
	outfile = fopen(path,"w");
	for (int k=0; k<channels; k++) {
		sum[k]=0;
		coef[k]=1;
	}
	if (normalize == true) {
		for (int i=0; i<y_bins; i++) {
			for (int k=0; k<channels; k++) {
				for (int n = -w; n <= w; n++) {
					sum[k]+=hist[i][y+n][k];
				}
			}
		}
		for (int k=0; k<channels; k++) {
			coef[k]=1e6/sum[k];
		}
	}
	for (int i=0; i<x_bins; i++) {
		fprintf(outfile, "%f", x_lower + i * 1.0 / (x_bins - 1) * (x_upper - x_lower));
		for (int k=0; k<channels; k++) {
			sum[k]=0;
			for (int n = -w; n <= w; n++) {
				sum[k]+=hist[i][y+n][k]*coef[k];
			}
			fprintf(outfile, " %f", sum[k]);
		}
		fprintf(outfile, "\n");			
	}	
	fclose(outfile);	
}

void histogram2D::writeImage(const char path[], const bool logscale)
{
// 	struct color palette[6];
// 
// 	(palette[5]).r=255;
// 	(palette[5]).g=0;
// 	(palette[5]).b=0;
// 
// 	(palette[4]).r=255;
// 	(palette[4]).g=255;
// 	(palette[4]).b=0;
// 
// 	(palette[3]).r=0;
// 	(palette[3]).g=255;
// 	(palette[3]).b=0;
// 
// 	(palette[2]).r=0;
// 	(palette[2]).g=255;
// 	(palette[2]).b=255;
// 
// 	(palette[1]).r=0;
// 	(palette[1]).g=0;
// 	(palette[1]).b=255;
// 
// 	(palette[0]).r=64;
// 	(palette[0]).g=0;
// 	(palette[0]).b=64;

	struct color palette[12];

// 	(palette[0]).r=0;
// 	(palette[0]).g=0;
// 	(palette[0]).b=0;
// 
// 	(palette[1]).r=204;
// 	(palette[1]).g=0;
// 	(palette[1]).b=115;
// 
// 	(palette[2]).r=112;
// 	(palette[2]).g=10;
// 	(palette[2]).b=171;
// 
// 	(palette[3]).r=56;
// 	(palette[3]).g=20;
// 	(palette[3]).b=176;
// 
// 	(palette[4]).r=18;
// 	(palette[4]).g=64;
// 	(palette[4]).b=171;
// 	
// 	(palette[5]).r=0;
// 	(palette[5]).g=153;
// 	(palette[5]).b=153;
// 
// 	(palette[6]).r=0;
// 	(palette[6]).g=204;
// 	(palette[6]).b=0;
// 
// 	(palette[7]).r=158;
// 	(palette[7]).g=237;
// 	(palette[7]).b=0;
// 
// 	(palette[8]).r=255;
// 	(palette[8]).g=255;
// 	(palette[8]).b=0;
// 
// 	(palette[9]).r=255;
// 	(palette[9]).g=212;
// 	(palette[9]).b=0;
// 
// 	(palette[10]).r=255;
// 	(palette[10]).g=171;
// 	(palette[10]).b=0;
// 
// 	(palette[11]).r=255;
// 	(palette[11]).g=115;
// 	(palette[11]).b=0;
// 
// 	(palette[12]).r=255;
// 	(palette[12]).g=0;
// 	(palette[12]).b=0;

	(palette[0]).r=0;
	(palette[0]).g=0;
	(palette[0]).b=0;

	(palette[1]).r=112;
	(palette[1]).g=10;
	(palette[1]).b=171;

	(palette[2]).r=56;
	(palette[2]).g=20;
	(palette[2]).b=176;

	(palette[3]).r=18;
	(palette[3]).g=64;
	(palette[3]).b=171;
	
	(palette[4]).r=0;
	(palette[4]).g=153;
	(palette[4]).b=153;

	(palette[5]).r=0;
	(palette[5]).g=204;
	(palette[5]).b=0;

	(palette[6]).r=158;
	(palette[6]).g=237;
	(palette[6]).b=0;

	(palette[7]).r=255;
	(palette[7]).g=255;
	(palette[7]).b=0;

	(palette[8]).r=255;
	(palette[8]).g=212;
	(palette[8]).b=0;

	(palette[9]).r=255;
	(palette[9]).g=171;
	(palette[9]).b=0;

	(palette[10]).r=255;
	(palette[10]).g=115;
	(palette[10]).b=0;

	(palette[11]).r=255;
	(palette[11]).g=0;
	(palette[11]).b=0;

	FILE *outfile;
	char path_suff[256];
	
	double min, max;
	int r, g, b;
	double shift, normalization;
	

	
	for (int k=0; k<channels; k++) {
		sprintf(path_suff, "%s%d.ppm",path,k+1);
		outfile = fopen(path_suff,"w");
		fprintf(outfile, "P3\n");
		fprintf(outfile, "%d %d\n", x_bins, y_bins);
		fprintf(outfile, "%d\n", 255);
		
		min=1e32;		
		max=hist[1][1][k];
		for (int i = 1; i < x_bins-1; i++) {
			for (int j = 1; j < y_bins-1; j++) {
				if (hist[i][j][k] > max) max = hist[i][j][k];
				if ((hist[i][j][k] < min) && (hist[i][j][k] > 0)) min = hist[i][j][k];				
			}
		}
		shift = 0;
		normalization = 1;
		if ((logscale == true) && (max > 0) && (min > 0)) {
			shift = log(0.5*min);
			if (logscale == true) normalization = 1.0/(log(max)-shift);
			else normalization = 1.0/max;
		}
		printf("Shift %g, normalization %g\n", shift, normalization);
		for (int j=y_bins-1; j>=0; j--) {		
			for (int i=0; i<x_bins; i++) {
// 				if (i == 140) hist[i][j][k] = 0.01;
// 				if (i == 145) hist[i][j][k] = 0.1;
// 				if (i == 150) hist[i][j][k] = 1;
// 				if (i == 155) hist[i][j][k] = 10;
// 				if (i == 160) hist[i][j][k] = 100;
// 				if (i == 165) hist[i][j][k] = 1000;
// 				if (i == 170) hist[i][j][k] = 10000;
				if (logscale == true) {
					if (hist[i][j][k] == 0) interpolate(hist[i][j][k]*normalization, &r, &g, &b, palette, 12);
					else interpolate((log(hist[i][j][k]) - shift)*normalization, &r, &g, &b, palette, 12);
				}
				else interpolate(hist[i][j][k]*normalization, &r, &g, &b, palette, 12);
				fprintf(outfile, "%d %d %d\n", r, g, b);	
			}
		}
		fclose(outfile);	
	}

}