/*
 *  Copyright (C) 2012  Marian Vlcek, Jakub Cizek
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *  If you publish any work which uses this program and/or data
 *  produced by this program, please cite following paper: 
 * 
 *  J. Cizek. M. Vlcek, I. Prochazka, Digital spectrometer
 *  for coincidence measurement of Doppler broadening of positron
 *  annihilation radiation, Nuclear Instruments and Methods
 *  in Physics Research, Section A 623, 982-994 (2010)
 */

#include "stdafx.h"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "histogram.h"

#define PI 3.1415926535897932384626433

#define WAVEFORMS_PER_FILE 10000
#define SESSIONS 1
#define SAMPLES 1000
#define NBR_PARAMETERS 4
#define DATA_OFFSET 12
#define NOISE_LEVEL 20

// g++ sdcdb_gen.cpp histogram.cpp -O2 -o sdcdb_gen
// ./sdcdb_gen

histogram1D hist(1601, 2, -8, 8);

int generator_state=1;

double normal_value=0;
int normal_state=0;


FILE *cdbfile;
FILE *sglfile;
FILE *examplefile;

int random_int() {
	generator_state=((long long) generator_state*69621) % 2147483647;
	return generator_state;
}

double uniform() {
	return random_int()*1.0/2147483647.0;
}

double normal() {
	if(normal_state==0) {
		double u1, u2, tmp;
		normal_state=1;
		u1=uniform();
		u2=uniform();
		tmp=sqrt(-2.0*log(u1));
		normal_value=tmp*cos(2.0*PI*u2);
		return tmp*sin(2.0*PI*u2);
	} else {
		normal_state=0;
		return normal_value;
	}
}

double f(double x, double par[NBR_PARAMETERS])
{
	return par[0]*exp(-(x-par[1])*(x-par[1])/(par[2]*par[2])) + par[3];
}

void generate_params(double par[NBR_PARAMETERS]) {
	double val;
	val=uniform();
	if(val < 0.3) {
		par[0]=12775*(1+0.003*normal());
	} else if(val < 0.5) {
		par[0]=31863*(1+0.0015*normal());
	} else {
		par[0]=10000+25000*uniform();
	}
	
	par[1]=600*(1+0.005*normal());
	par[2]=175*(1+0.0005*normal());
	par[3]=-28000*(1+0.05*normal());
}

int main(int argc, char *argv[])
{
	
	int i, n, k;
	char filename[256];
	double par1[NBR_PARAMETERS];
	double par2[NBR_PARAMETERS];
	short short1, short2;
	char pad;
	
	printf("SDCDB Hist  Copyright (C) 2012  Marian Vlcek, Jakub Cizek\n");
	
	generator_state=1;
	printf("Testing random number generator...\n");
	for(i=0; i<1000000; i++) {
		hist.add(uniform(), 0);
		hist.add(normal(), 1);
	}
	hist.write("sdcdb_gen_test.txt");
	
	generator_state=1;
	examplefile=fopen("sdcdb_gen_example.txt","w");
	generate_params(par1);
	generate_params(par2);
	for(i=0; i<SAMPLES; i++) {
		short1=(short) (f(i, par1) + NOISE_LEVEL*normal());
		short2=(short) (f(i, par2) + NOISE_LEVEL*normal());
		fprintf(examplefile, "%d %d %d\n", i, short1, short2);
	}
	fclose(examplefile);
	
	generator_state=1;
	printf("Generating synthetic data:\n");
	for(k=1; k<=SESSIONS; k++) {
		sprintf(filename, "synth-%d.cdb", k);
		printf("%s, ", filename);
		cdbfile=fopen(filename, "wb");
		sprintf(filename, "synth-%d.sgl", k);
		printf("%s\n", filename);
		sglfile=fopen(filename, "wb");
		
		pad=0;
		for(i=0; i<DATA_OFFSET; i++) {
			fwrite(&pad, 1, sizeof(pad), cdbfile);
			fwrite(&pad, 1, sizeof(pad), sglfile);
		}
		
		for(n=0; n<WAVEFORMS_PER_FILE; n++) {
			generate_params(par1);
			generate_params(par2);
			
			for(i=0; i<SAMPLES; i++) {
				short1=(short) (f(i, par1) + NOISE_LEVEL*normal());
				short2=(short) (f(i, par2) + NOISE_LEVEL*normal());
				
				fwrite(&short1, 1, sizeof(short), cdbfile);
				fwrite(&short2, 1, sizeof(short), cdbfile);
				
				fwrite(&short1, 1, sizeof(short), sglfile);
				fwrite(&short2, 1, sizeof(short), sglfile);
			}
		}
		fclose(cdbfile);
		fclose(sglfile);
	}
	
	return 0;
}

/*
#include "stdafx.h"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "histogram.h"

#define PI 3.1415926535897932384626433

histogram1D hist(1601, 2, -8, 8);

int generator_state=1;

double normal_value=0;
int normal_state=0;

FILE *outfile;

int random_int() {
	generator_state=((long long) generator_state*69621) % 2147483647;
	return generator_state;
}

double uniform() {
	return random_int()*1.0/2147483647.0;
}

double normal() {
	if(normal_state==0) {
		double u1, u2, tmp;
		normal_state=1;
		u1=uniform();
		u2=uniform();
		tmp=sqrt(-2.0*log(u1));
		normal_value=tmp*cos(2.0*PI*u2);
		return tmp*sin(2.0*PI*u2);
	} else {
		normal_state=0;
		return normal_value;
	}
}



int main(int argc, char *argv[])
{
	
	int i, n, k;
	char filename[256];
	double par1[4];
	double par2[4];
	short short1, short2;
	
	generator_state=1;
	printf("Testing random number generator...\n");
	for(i=0; i<1000000; i++) {
		hist.add(uniform(), 0);
		hist.add(normal(), 1);
	}
	hist.write("dcdb_gen_test.txt");
	
	generator_state=1;
	printf("Generating synthetic data:\n");
// 	for(k=1; k<=10; k++) {
// 		sprintf(filename, "synth.%d", k);
// 		printf("%s\n", filename);
// 		outfile=fopen(filename, "wb");
// 		n=0;
// 		fwrite(&n, 1, sizeof(n), outfile);
// 		fwrite(&n, 1, sizeof(n), outfile);
// 		fwrite(&n, 1, sizeof(n), outfile);
// 		for(n=0; n<10000; n++) {
// 			generate_params(par1);
// 			generate_params(par2);
// 			
// 			for(i=0; i<2500; i++) {
// 				short1=(short) (f(i, par1) + 50*normal());
// 				short2=(short) (f(i, par2) + 50*normal());
// 				fwrite(&short1, 1, sizeof(short), outfile);
// 				fwrite(&short2, 1, sizeof(short), outfile);
// 			}
// 		}
// 		fclose(outfile);
// 	}
	
	FILE *testfile;
	testfile=fopen("test.txt","w");
	generate_params(par1);
	generate_params(par2);
	for(i=0; i<1000; i++) {
		short1=(short) (f(i, par1) + 10*normal());
		short2=(short) (f(i, par2) + 10*normal());
		fprintf(testfile, "%d %d %d\n", i, short1, short2);
	}
	fclose(testfile);
	
	return 0;
}*/